package ru.dbotthepony.mc.otm.block.entity.blackhole

import net.minecraft.client.Minecraft
import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.EquipmentSlot
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.item.ItemEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.Explosion
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.levelgen.structure.BoundingBox
import net.minecraft.world.phys.AABB
import net.minecraft.world.phys.Vec3
import ru.dbotthepony.mc.otm.block.BlackHoleBlock
import ru.dbotthepony.mc.otm.block.entity.GravitationStabilizerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.SynchronizedBlockEntity
import ru.dbotthepony.mc.otm.block.entity.blackhole.ExplosionQueue.Companion.queueForLevel
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.plus
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MItems
import ru.dbotthepony.mc.otm.registry.MRegistry
import ru.dbotthepony.mc.otm.core.getSphericalBlockPositions
import ru.dbotthepony.mc.otm.core.mapIf
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.matter.MatterManager
import ru.dbotthepony.mc.otm.triggers.BlackHoleTrigger
import java.util.LinkedList
import kotlin.math.pow
import kotlin.math.roundToInt
import kotlin.math.sqrt

class BlackHoleBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) : SynchronizedBlockEntity(MBlockEntities.BLACK_HOLE, p_155229_, p_155230_) {
	var mass by synchronizer.fraction(BASELINE_MASS, setter = setter@{ mass, field, setByRemote ->
		if (mass <= ImpreciseFraction.ZERO) {
			collapse()
			return@setter
		}

		field.write(mass)
		setChanged()
		var massForDiv = mass

		when (if (level?.isClientSide == false) stabilizers.size else stabilizerClientCount) {
			1 -> {massForDiv /= 4}
			2 -> {massForDiv /= 16}
			3 -> {massForDiv /= 64}
			4 -> {massForDiv /= 512}
			5 -> {massForDiv /= 2048}
			6 -> {massForDiv /= 16384}
		}

		gravitationStrength = sqrt(massForDiv.div(BASELINE_MASS).toDouble()).coerceAtMost(20.0).coerceAtLeast(0.2)

		affectedBounds = BoundingBox(
			(-30 * gravitationStrength).toInt(),
			(-30 * gravitationStrength).toInt(),
			(-30 * gravitationStrength).toInt(),
			(30 * gravitationStrength).toInt(),
			(30 * gravitationStrength).toInt(),
			(30 * gravitationStrength).toInt()
		).moved(
			blockPos.x, blockPos.y, blockPos.z
		)

		affectedBoundsAABB = AABB.of(affectedBounds)
	})

	override fun getRenderBoundingBox(): AABB {
		return AABB(blockPos.offset(-GravitationStabilizerBlockEntity.RANGE, -GravitationStabilizerBlockEntity.RANGE, -GravitationStabilizerBlockEntity.RANGE), blockPos.offset(GravitationStabilizerBlockEntity.RANGE, GravitationStabilizerBlockEntity.RANGE, GravitationStabilizerBlockEntity.RANGE))
	}

	var gravitationStrength = 1.0
		private set

	private var suppressUpdates = true
	var spinDirection = false

	private var sleepTicks = 4

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		sleepTicks = 4
	}

	private fun updateMass() {
		mass = mass
	}

	private val stabilizers = LinkedList<GravitationStabilizerBlockEntity>()
	private var stabilizerClientCount by synchronizer.int(setter = setter@{ value, field, setByRemote ->
		field.write(value)

		if (setByRemote) {
			updateMass()
		}
	})

	fun stabilizerAttached(stabilizer: GravitationStabilizerBlockEntity) {
		if (stabilizer in stabilizers)
			return

		stabilizers.add(stabilizer)
		mass = mass
		stabilizerClientCount = stabilizers.size
	}

	fun stabilizerDetached(stabilizer: GravitationStabilizerBlockEntity) {
		if (stabilizers.remove(stabilizer)) {
			mass = mass
			stabilizerClientCount = stabilizers.size
		}
	}

	fun collapse() {
		val level = level as? ServerLevel ?: return

		level.setBlock(blockPos, Blocks.AIR.defaultBlockState(), Block.UPDATE_ALL)

		if (gravitationStrength > 0.25) {
			val x0 = blockPos.x + 0.5
			val y0 = blockPos.y + 0.5
			val z0 = blockPos.z + 0.5
			val queue = queueForLevel(level)
			var radius = 0

			while (radius < Math.ceil(gravitationStrength * 4)) {
				queue.explodeRing(
					x0,
					y0,
					z0,
					radius.toDouble(), Math.min(20.0, Math.max(1.0, (gravitationStrength * 4 - radius) * 20)).toFloat()
				)

				radius++
			}
		} else {
			level.explode(
				null,
				MRegistry.DAMAGE_HAWKING_RADIATION,
				null,
				blockPos.x + 0.5,
				blockPos.y + 0.5,
				blockPos.z + 0.5,
				gravitationStrength.toFloat() * 60,
				false,
				Explosion.BlockInteraction.DESTROY
			)
		}
	}

	public override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["mass"] = mass.serializeNBT()
		nbt["spin_direction"] = spinDirection
	}

	override fun load(tag: CompoundTag) {
		super.load(tag)
		mass = tag.mapIf("mass", ImpreciseFraction::deserializeNBT) ?: BASELINE_MASS
		spinDirection = tag.getBoolean("spin_direction")
	}

	var affectedBounds = BoundingBox(0, 0, 0, 1, 1, 1)
		private set
	var affectedBoundsAABB: AABB = AABB.of(affectedBounds)
		private set

	private fun setDeltaMovement(living: Entity, center: Vec3, distance: Double, weaker: Boolean) {
		//final double mult = Math.min(2, (30 * this.gravitation_strength) / Math.max(1, Math.pow(distance, 2)));
		// Сила притяжения
		val mult = Math.pow((1 - distance / (30 * gravitationStrength)).coerceAtLeast(0.0), 2.0) * gravitationStrength / 8

		// Притяжение к ядру
		val delta = living.position().vectorTo(center).normalize()

		if (distance < gravitationStrength) {
			living.deltaMovement = living.deltaMovement.add(delta.multiply(mult, mult, mult))
		} else {
			// Закручивание
			val rotate = if (spinDirection) delta.yRot((Math.PI / 2).toFloat()) else delta.yRot((-Math.PI / 2).toFloat())

			if (weaker)
				living.deltaMovement = living.deltaMovement
					.add(rotate.multiply(mult * 0.2f, mult * 0.2f, mult * 0.2f))
					.add(delta.multiply(mult * 0.33f, mult * 0.33f, mult * 0.33f))
			else
				living.deltaMovement = living.deltaMovement
					.add(rotate.multiply(mult * 0.2f, mult * 0.2f, mult * 0.2f))
					.add(delta.multiply(mult, mult, mult))
		}
	}

	fun clientTick() {
		sleepTicks--
		if (sleepTicks > 0) return
		val ply = Minecraft.getInstance().player!!
		val center = Vec3.atCenterOf(blockPos)

		if (!ply.abilities.mayfly && ply.getItemBySlot(EquipmentSlot.CHEST).item != MItems.PORTABLE_GRAVITATION_STABILIZER) {
			val distance = ply.position().distanceTo(center)
			setDeltaMovement(ply, center, distance, true)
		}

		for (item in level!!.getEntitiesOfClass(ItemEntity::class.java, affectedBoundsAABB)) {
			val distance = item.position().distanceTo(center)
			setDeltaMovement(item, center, distance, false)
		}
	}

	fun tick() {
		sleepTicks--
		if (sleepTicks > 0) return
		val level = level as? ServerLevel ?: return
		suppressUpdates = false

		val center = Vec3.atCenterOf(blockPos)

		kotlin.run {
			val gravitationStrength = gravitationStrength.pow(0.5)

			for (living in level.getEntitiesOfClass(LivingEntity::class.java, affectedBoundsAABB)) {
				val distance = living.position().distanceTo(center)

				if (living !is Player || !living.abilities.mayfly && living.getItemBySlot(EquipmentSlot.CHEST).item != MItems.PORTABLE_GRAVITATION_STABILIZER) {
					setDeltaMovement(living, center, distance, living !is Player)

					if (living is ServerPlayer) {
						BlackHoleTrigger.trigger(living)
					}
				}

				if (distance < gravitationStrength + 1) {
					living.hurt(MRegistry.DAMAGE_EVENT_HORIZON, (gravitationStrength / distance).toFloat())
				}
			}

			for (item in level.getEntitiesOfClass(ItemEntity::class.java, affectedBoundsAABB)) {
				val distance = item.position().distanceTo(center)
				setDeltaMovement(item, center, distance, false)

				if (distance < gravitationStrength + 1) {
					if (item.hurt(MRegistry.DAMAGE_EVENT_HORIZON, (gravitationStrength / distance).toFloat()) && item.isRemoved) {
						if (item.item.item === MItems.GRAVITATIONAL_DISRUPTOR) {
							collapse()
						} else {
							val mass = MatterManager.get(item.item)

							if (mass.hasMatterValue)
								this.mass += mass.matter
						}
					}
				}
			}
		}

		// шанс 1% что черная дыра потеряет 0.1 MtU каждую секунду * силу гравитации дыры ^ -1
		if (level.random.nextDouble() < 0.01 * 0.05 * (1 / gravitationStrength)) {
			this.mass += HAWKING_MASS_LOSE_STEP
		}

		if (gravitationStrength > 0.4) {
			val sphere = getSphericalBlockPositions((gravitationStrength * 6.0).roundToInt())

			if (sphere.size != lastSphereSizeOuter) {
				lastSphereSizeOuter = sphere.size
				sphereIndexOuter = 0
			}

			var iterations = 0

			while (iterations < ITERATIONS) {
				iterations++

				val pos = sphere[sphereIndexOuter] + blockPos
				sphereIndexOuter = (sphereIndexOuter + 1) % lastSphereSizeOuter

				if (sphereIndexOuter == 0 && lastSphereSizeOuter <= ITERATIONS)
					break

				val getBlock = level.getBlockState(pos)

				if (!getBlock.isAir && getBlock.block !is BlackHoleBlock) {
					val speed = getBlock.getDestroySpeed(level, pos)

					val eResist = try {
						getBlock.getExplosionResistance(level, pos, null)
					} catch (err: NullPointerException) {
						getBlock.block.explosionResistance
						// Потому что возможно какой-либо мод не ожидает что Explosion == null
						// особенно учитывая что интерфейс IForgeBlock не имеет @ParamsAreNonnullByDefault
						// и аргумент не помечен как @Nullable
						// тем самым имеет тип Explosion! который указывается как Explosion? .. Explosion!!
					} catch (err: IllegalArgumentException) {
						getBlock.block.explosionResistance
					}

					var strengthLinear = sqrt(blockPos.distSqr(pos))

					if (strengthLinear < gravitationStrength * 1.1) {
						strengthLinear /= 4.0
					}

					if (speed >= 0f && (speed * 2f).coerceAtLeast(eResist / 3f) < gravitationStrength * (8.0 / strengthLinear)) {
						Block.dropResources(getBlock, level, pos, level.getBlockEntity(pos))
						getBlock.block.destroy(level, pos, getBlock)
						level.setBlock(pos, getBlock.fluidState.createLegacyBlock(), Block.UPDATE_ALL)
					}
				}
			}
		}
	}

	private var lastSphereSizeOuter = 0
	private var sphereIndexOuter = 0

	init {
		mass = mass
	}

	companion object {
		const val ITERATIONS = 30_000
		val BASELINE_MASS = ImpreciseFraction(1_000)
		val HAWKING_MASS_LOSE_STEP = ImpreciseFraction("-0.1")
	}
}
