package ru.dbotthepony.mc.otm.capability

import com.google.common.collect.Streams
import net.minecraft.core.Direction
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.IEnergyStorage
import net.minecraftforge.fml.ModList
import ru.dbotthepony.mc.otm.compat.cos.cosmeticArmorAwareStream
import ru.dbotthepony.mc.otm.compat.cos.cosmeticArmorStream
import ru.dbotthepony.mc.otm.compat.cos.isCosmeticArmorLoaded
import ru.dbotthepony.mc.otm.compat.curios.curiosAwareStream
import ru.dbotthepony.mc.otm.compat.curios.curiosStream
import ru.dbotthepony.mc.otm.compat.curios.isCuriosLoaded
import ru.dbotthepony.mc.otm.compat.mekanism.getMekanismEnergySided
import ru.dbotthepony.mc.otm.compat.mekanism.mekanismEnergy
import ru.dbotthepony.mc.otm.container.awareStream
import ru.dbotthepony.mc.otm.container.stream
import ru.dbotthepony.mc.otm.core.AwareItemStack
import ru.dbotthepony.mc.otm.core.ContainerItemStackEntry
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.orNull
import java.util.IdentityHashMap
import java.util.LinkedList
import java.util.stream.Stream

val ICapabilityProvider.matteryPlayer: MatteryPlayerCapability? get() = getCapability(MatteryCapability.MATTERY_PLAYER).orNull()

fun IEnergyStorage.receiveEnergy(amount: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	if (this is IMatteryEnergyStorage)
		return receiveEnergyOuter(amount, simulate)

	if (!amount.isPositive)
		return ImpreciseFraction.ZERO

	if (amount > ImpreciseFraction.INT_MAX_VALUE)
		return ImpreciseFraction.valueOf(receiveEnergy(Int.MAX_VALUE, simulate))

	return ImpreciseFraction.valueOf(receiveEnergy(amount.toInt(), simulate))
}

fun IEnergyStorage.extractEnergy(amount: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	if (this is IMatteryEnergyStorage)
		return extractEnergyOuter(amount, simulate)

	if (!amount.isPositive)
		return ImpreciseFraction.ZERO

	if (amount > ImpreciseFraction.INT_MAX_VALUE)
		return ImpreciseFraction.valueOf(extractEnergy(Int.MAX_VALUE, simulate))

	return ImpreciseFraction.valueOf(extractEnergy(amount.toInt(), simulate))
}

val IEnergyStorage.maxEnergyStoredMattery: ImpreciseFraction get() {
	if (this is IMatteryEnergyStorage) {
		return maxBatteryLevel
	}

	return ImpreciseFraction.valueOf(maxEnergyStored)
}

val IEnergyStorage.energyStoredMattery: ImpreciseFraction get() {
	if (this is IMatteryEnergyStorage) {
		return batteryLevel
	}

	return ImpreciseFraction.valueOf(energyStored)
}

val IEnergyStorage.chargeRatio: Float get() {
	if (this is IMatteryEnergyStorage) {
		return (batteryLevel / maxBatteryLevel).toFloat()
	}

	return energyStored.toFloat() / maxEnergyStored.toFloat()
}

val isMekanismLoaded by lazy { ModList.get().isLoaded("mekanism") }

/**
 * Shortcut for getting [IEnergyStorage], including wrappers for it
 */
val ICapabilityProvider.energy: IEnergyStorage? get() {
	val mattery = getCapability(MatteryCapability.ENERGY)

	if (mattery.isPresent) {
		return mattery.orNull()
	}

	if (isMekanismLoaded) {
		val mekanismEnergy = mekanismEnergy

		if (mekanismEnergy != null) {
			return mekanismEnergy
		}
	}

	return getCapability(ForgeCapabilities.ENERGY).orNull()
}

/**
 * Shortcut for getting [IMatteryEnergyStorage], including wrappers for it
 */
val ICapabilityProvider.matteryEnergy: IMatteryEnergyStorage? get() {
	val mattery = getCapability(MatteryCapability.ENERGY)

	if (mattery.isPresent) {
		return mattery.orNull()
	}

	if (isMekanismLoaded) {
		val mekanismEnergy = mekanismEnergy

		if (mekanismEnergy != null) {
			return mekanismEnergy
		}
	}

	return null
}

/**
 * Shortcut for getting [LazyOptional] with [IEnergyStorage], including wrappers for it
 */
fun ICapabilityProvider.getEnergySided(side: Direction? = null): LazyOptional<out IEnergyStorage> {
	val mattery = getCapability(MatteryCapability.ENERGY, side)

	if (mattery.isPresent) {
		return mattery
	}

	if (isMekanismLoaded) {
		val mekanismEnergy = getMekanismEnergySided(side)

		if (mekanismEnergy.isPresent) {
			return mekanismEnergy
		}
	}

	return getCapability(ForgeCapabilities.ENERGY, side)
}

/**
 * Shortcut for getting [LazyOptional] with [IMatteryEnergyStorage], including wrappers for it
 */
fun ICapabilityProvider.getMatteryEnergySided(side: Direction? = null): LazyOptional<out IMatteryEnergyStorage> {
	val mattery = getCapability(MatteryCapability.ENERGY, side)

	if (mattery.isPresent) {
		return mattery
	}

	if (isMekanismLoaded) {
		val mekanismEnergy = getMekanismEnergySided(side)

		if (mekanismEnergy.isPresent) {
			return mekanismEnergy
		}
	}

	return LazyOptional.empty()
}

/**
 * DO NOT modify returned ItemStacks!
 *
 * Contains all items that player might carry
 */
fun Player.itemsStream(includeCosmetics: Boolean = true): Stream<out ItemStack> {
	val streams = LinkedList<Stream<out ItemStack>>()
	streams.add(inventory.stream())

	matteryPlayer?.let {
		if (it.hasExoSuit) {
			streams.add(it.exoSuitContainer.stream())
		}
	}

	if (isCuriosLoaded) {
		streams.add(curiosStream(includeCosmetics))
	}

	if (includeCosmetics && isCosmeticArmorLoaded) {
		streams.add(cosmeticArmorStream())
	}

	return Streams.concat(*streams.toTypedArray())
}

/**
 * DO NOT modify returned ItemStacks!
 *
 * Contains all items that player might see/access ([itemsStream] + open container's items)
 */
fun Player.allItemsStream(includeCosmetics: Boolean = true): Stream<out ItemStack> {
	if (containerMenu == inventoryMenu || containerMenu == matteryPlayer?.exoSuitMenu) {
		return itemsStream(includeCosmetics)
	}

	val seen = IdentityHashMap<ItemStack, Unit>(containerMenu.slots.size)

	for (slot in containerMenu.slots) {
		seen[slot.item] = Unit
	}

	return Streams.concat(
		itemsStream(includeCosmetics).filter { it.isEmpty || it !in seen },
		containerMenu.slots.stream().map { it.item })
}

/**
 * Modify returned [ItemStack]s through [AwareItemStack.extract]
 *
 * Contains all items that player might carry
 */
fun Player.awareItemsStream(includeCosmetics: Boolean = false): Stream<out AwareItemStack> {
	val streams = LinkedList<Stream<out AwareItemStack>>()
	streams.add(inventory.awareStream())

	matteryPlayer?.let {
		if (it.hasExoSuit) {
			streams.add(it.exoSuitContainer.awareStream())
		}
	}

	if (isCuriosLoaded) {
		streams.add(curiosAwareStream(includeCosmetics))
	}

	if (includeCosmetics && isCosmeticArmorLoaded) {
		streams.add(cosmeticArmorAwareStream())
	}

	return Streams.concat(*streams.toTypedArray())
}

/**
 * Modify returned [ItemStack]s through [AwareItemStack.extract]
 *
 * Contains all items that player might see/access ([itemsStream] + open container's items)
 */
fun Player.awareAllItemsStream(includeCosmetics: Boolean = false): Stream<out AwareItemStack> {
	if (containerMenu == inventoryMenu || containerMenu == matteryPlayer?.exoSuitMenu) {
		return awareItemsStream(includeCosmetics)
	}

	val seen = HashSet<AwareItemStack>(containerMenu.slots.size)

	for (slot in containerMenu.slots) {
		seen.add(ContainerItemStackEntry(slot.slotIndex, slot.container))
	}

	return Streams.concat(
		awareItemsStream(includeCosmetics).filter { it !in seen },
		seen.stream())
}

