package ru.dbotthepony.mc.otm.capability

import net.minecraft.util.Mth
import net.minecraftforge.energy.IEnergyStorage
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.RGBAColor
import java.math.BigInteger
import kotlin.math.roundToInt

// IEnergyStorage for direct compat with Forge Energy
interface IMatteryEnergyStorage : IEnergyStorage {
	/**
	 * Energy extraction by external interacts
	 *
	 * @return energy extracted
	 */
	fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction

	/**
	 * Energy extraction by internal processes
	 *
	 * Nothing stops you from calling this directly, not on yours capability,
	 * but you really should think before doing this on not own capability.
	 *
	 * @return energy extracted
	 */
	fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction

	/**
	 * Energy insertion by external interacts
	 *
	 * @return energy accepted
	 */
	fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction

	/**
	 * Energy insertion by internal processes
	 *
	 * Nothing stops you from calling this directly, not on yours capability,
	 * but you really should think before doing this on not own capability.
	 *
	 * @return energy accepted
	 */
	fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction

	val batteryLevel: ImpreciseFraction
	val maxBatteryLevel: ImpreciseFraction
	val missingPower: ImpreciseFraction
		get() = (maxBatteryLevel - batteryLevel).moreThanZero()

	override fun receiveEnergy(maxReceive: Int, simulate: Boolean): Int {
		val received = receiveEnergyOuter(maxReceive, true).toInt()

		// Receiving only a fraction
		if (received == 0)
			return 0

		return receiveEnergyOuter(ImpreciseFraction(received), simulate).toInt()
	}

	override fun extractEnergy(maxReceive: Int, simulate: Boolean): Int {
		val extracted = extractEnergyOuter(maxReceive, true).toInt()

		// Extracting only a fraction
		if (extracted == 0)
			return 0

		return extractEnergyOuter(ImpreciseFraction(extracted), simulate).toInt()
	}

	override fun getEnergyStored(): Int {
		return batteryLevel.toInt()
	}

	override fun getMaxEnergyStored(): Int {
		return maxBatteryLevel.toInt()
	}

	override fun canExtract(): Boolean {
		return extractEnergyOuter(ImpreciseFraction.ONE, true) > ImpreciseFraction.ZERO
	}

	override fun canReceive(): Boolean {
		return receiveEnergyOuter(ImpreciseFraction.ONE, true) > ImpreciseFraction.ZERO
	}
}

fun IMatteryEnergyStorage.receiveEnergyOuterExact(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	val extracted = receiveEnergyOuter(howMuch, true)

	if (extracted != howMuch) {
		return ImpreciseFraction.ZERO
	}

	if (!simulate) {
		receiveEnergyOuter(howMuch, false)
	}

	return extracted
}
/**
 * All or nothing
 *
 * @return energy extracted
 */
fun IMatteryEnergyStorage.extractEnergyOuterExact(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	val extracted = extractEnergyOuter(howMuch, true)

	if (extracted != howMuch) {
		return ImpreciseFraction.ZERO
	}

	if (!simulate) {
		extractEnergyOuter(howMuch, false)
	}

	return extracted
}

/**
 * All or nothing
 *
 * @return energy accepted
 */
fun IMatteryEnergyStorage.receiveEnergyInnerExact(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	val extracted = receiveEnergyInner(howMuch, true)

	if (extracted != howMuch) {
		return ImpreciseFraction.ZERO
	}

	if (!simulate) {
		receiveEnergyInner(howMuch, false)
	}

	return extracted
}

/**
 * All or nothing
 *
 * @return energy extracted
 */
fun IMatteryEnergyStorage.extractEnergyInnerExact(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	val extracted = extractEnergyInner(howMuch, true)

	if (extracted != howMuch) {
		return ImpreciseFraction.ZERO
	}

	if (!simulate) {
		extractEnergyInner(howMuch, false)
	}

	return extracted
}

fun IMatteryEnergyStorage.extractEnergyOuter(howMuch: Long, simulate: Boolean): ImpreciseFraction {
	return extractEnergyOuter(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.extractEnergyOuter(howMuch: Int, simulate: Boolean): ImpreciseFraction {
	return extractEnergyOuter(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.receiveEnergyOuter(howMuch: Long, simulate: Boolean): ImpreciseFraction {
	return receiveEnergyOuter(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.receiveEnergyOuter(howMuch: Int, simulate: Boolean): ImpreciseFraction {
	return receiveEnergyOuter(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.extractEnergyInner(howMuch: Long, simulate: Boolean): ImpreciseFraction {
	return extractEnergyInner(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.extractEnergyInner(howMuch: Int, simulate: Boolean): ImpreciseFraction {
	return extractEnergyInner(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.receiveEnergyInner(howMuch: Long, simulate: Boolean): ImpreciseFraction {
	return receiveEnergyInner(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.receiveEnergyInner(howMuch: Int, simulate: Boolean): ImpreciseFraction {
	return receiveEnergyInner(ImpreciseFraction(howMuch), simulate)
}

fun IMatteryEnergyStorage.transferInner(other: IMatteryEnergyStorage, amount: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	if (!amount.isPositive)
		return ImpreciseFraction.ZERO

	val extracted = extractEnergyInner(amount, true)
	val received = other.receiveEnergyOuter(extracted, simulate)

	if (!simulate)
		extractEnergyInner(received, false)

	return received
}

fun IMatteryEnergyStorage.transferOuter(other: IMatteryEnergyStorage, amount: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
	if (!amount.isPositive)
		return ImpreciseFraction.ZERO

	val extracted = extractEnergyOuter(amount, true)
	val received = other.receiveEnergyOuter(extracted, simulate)

	if (!simulate)
		extractEnergyInner(received, false)

	return received
}

fun IMatteryEnergyStorage.extractStepInner(base: ImpreciseFraction, multiplier: Int, simulate: Boolean): Int {
	return (extractEnergyInner(base * multiplier, simulate) / base).toInt()
}

fun IMatteryEnergyStorage.extractStepOuter(base: ImpreciseFraction, multiplier: Int, simulate: Boolean): Int {
	return (extractEnergyOuter(base * multiplier, simulate) / base).toInt()
}

fun IMatteryEnergyStorage.extractStepInnerBi(base: ImpreciseFraction, multiplier: Int, simulate: Boolean): BigInteger {
	return (extractEnergyInner(base * multiplier, simulate) / base).whole
}

fun IMatteryEnergyStorage.extractStepOuterBi(base: ImpreciseFraction, multiplier: Int, simulate: Boolean): BigInteger {
	return (extractEnergyOuter(base * multiplier, simulate) / base).whole
}

fun IMatteryEnergyStorage.extractStepInner(base: ImpreciseFraction, multiplier: BigInteger, simulate: Boolean): Int {
	return (extractEnergyInner(base * multiplier, simulate) / base).toInt()
}

fun IMatteryEnergyStorage.extractStepOuter(base: ImpreciseFraction, multiplier: BigInteger, simulate: Boolean): Int {
	return (extractEnergyOuter(base * multiplier, simulate) / base).toInt()
}

fun IMatteryEnergyStorage.extractStepInnerBi(base: ImpreciseFraction, multiplier: BigInteger, simulate: Boolean): BigInteger {
	return (extractEnergyInner(base * multiplier, simulate) / base).whole
}

fun IMatteryEnergyStorage.extractStepOuterBi(base: ImpreciseFraction, multiplier: BigInteger, simulate: Boolean): BigInteger {
	return (extractEnergyOuter(base * multiplier, simulate) / base).whole
}

fun IMatteryEnergyStorage.getBarWidth(): Int {
	return ((batteryLevel / maxBatteryLevel).toFloat().coerceAtLeast(0f).coerceAtMost(1f) * 13f).roundToInt()
}

fun IMatteryEnergyStorage.getBarColor(): Int {
	return RGBAColor.LOW_POWER.linearInterpolation((batteryLevel / maxBatteryLevel).toFloat(), RGBAColor.FULL_POWER).toInt()
}
