package ru.dbotthepony.mc.otm.data

import com.google.gson.Gson
import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonDeserializer
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSerializer
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import io.netty.buffer.ByteBufInputStream
import io.netty.buffer.ByteBufOutputStream
import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.core.readType
import ru.dbotthepony.mc.otm.core.readVarIntLE
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.writeType
import ru.dbotthepony.mc.otm.core.writeVarIntLE
import java.io.DataInputStream
import java.io.DataOutputStream
import java.lang.reflect.Type
import java.util.Base64
import java.util.Collections
import java.util.LinkedList
import java.util.function.Consumer

class SerializedFunctionRegistry<R, T>(val gson: Gson = Gson()) : JsonSerializer<SerializedFunctionRegistry.BoundFunction<R, T>>, JsonDeserializer<SerializedFunctionRegistry.BoundFunction<R, T>>, TypeAdapter<SerializedFunctionRegistry.BoundFunction<R, T>>() {
	fun interface AnonymousFunction<R, T> {
		fun invoke(receiver: R, arguments: List<Any?>): T
	}

	data class BoundFunction<R, T>(
		val function: Function<R, T>,
		val arguments: List<Any?>
	) : java.util.function.Function<R, T> {
		fun toJson(): JsonElement {
			return JsonObject().also {
				it["function"] = function.toJson()
				val stream = FastByteArrayOutputStream()
				val dataStream = DataOutputStream(stream)

				stream.writeVarIntLE(arguments.size)

				for (argument in arguments) {
					dataStream.writeType(argument)
				}

				it["arguments"] = JsonPrimitive(Base64.getEncoder().encode(stream.array.copyOfRange(0, stream.length)).toString(Charsets.UTF_8))
			}
		}

		fun toNetwork(buff: FriendlyByteBuf) {
			function.toNetwork(buff)

			val stream = DataOutputStream(ByteBufOutputStream(buff))
			stream.writeVarIntLE(arguments.size)
			for (argument in arguments) {
				stream.writeType(argument)
			}
		}

		override fun apply(t: R): T {
			return function.body.invoke(t, arguments)
		}
	}

	data class Function<R, T>(
		val id: ResourceLocation,
		val body: AnonymousFunction<R, T>,
		val registry: SerializedFunctionRegistry<R, T>
	) {
		fun bind(vararg arguments: Any?): BoundFunction<R, T> {
			validate(arguments.iterator().withIndex())
			return BoundFunction(this, Collections.unmodifiableList(LinkedList<Any?>().also { it.addAll(arguments) }))
		}

		fun bind(arguments: List<Any?>): BoundFunction<R, T> {
			validate(arguments.iterator().withIndex())
			return BoundFunction(this, Collections.unmodifiableList(LinkedList<Any?>().also { it.addAll(arguments) }))
		}

		fun toJson(): JsonElement {
			return JsonPrimitive(id.toString())
		}

		fun toNetwork(buff: FriendlyByteBuf) {
			buff.writeUtf(id.toString())
		}

		companion object {
			private fun validate(iterator: Iterator<IndexedValue<Any?>>) {
				try {
					val stream = DataOutputStream(FastByteArrayOutputStream())

					for ((i, argument) in iterator) {
						try {
							stream.writeType(argument)
						} catch(err: Exception) {
							throw IllegalArgumentException("Argument at $i can not be serialized", err)
						}
					}
				} catch(err: Exception) {
					throw IllegalArgumentException("Argument list validation failed", err)
				}
			}
		}
	}

	private val map = HashMap<ResourceLocation, Function<R, T>>()

	fun register(id: ResourceLocation, function: AnonymousFunction<R, T>): Function<R, T> {
		synchronized(map) {
			return map.computeIfAbsent(id) { Function(id, function, this) }
		}
	}

	fun register(id: ResourceLocation, function: R.() -> T): Function<R, T> {
		return register(id, AnonymousFunction { r, args ->
			check(args.isEmpty()) { "Invalid amount of arguments. No arguments are required" }
			function.invoke(r)
		})
	}

	inline fun <reified A : Any> register(id: ResourceLocation, noinline function: R.(A) -> T): Function<R, T> {
		return register(id, AnonymousFunction { r, args ->
			check(args.size == 1) { "Invalid amount of arguments. 1 is required" }
			function.invoke(
				r,
				args[0] as? A ?: throw ClassCastException("Argument at 0 is supposed to be ${A::class.qualifiedName}, ${args[0]?.let { it::class.qualifiedName }} given")
			)
		})
	}

	inline fun <reified A : Any, reified B : Any> register(id: ResourceLocation, noinline function: R.(A, B) -> T): Function<R, T> {
		return register(id, AnonymousFunction { r, args ->
			check(args.size == 2) { "Invalid amount of arguments. 2 is required" }

			function.invoke(
				r,
				args[0] as? A ?: throw ClassCastException("Argument at 0 is supposed to be ${A::class.qualifiedName}, ${args[0]?.let { it::class.qualifiedName }} given"),
				args[1] as? B ?: throw ClassCastException("Argument at 1 is supposed to be ${B::class.qualifiedName}, ${args[1]?.let { it::class.qualifiedName }} given"),
			)
		})
	}

	inline fun <reified A : Any, reified B : Any, reified C : Any> register(id: ResourceLocation, noinline function: R.(A, B, C) -> T): Function<R, T> {
		return register(id, AnonymousFunction { r, args ->
			check(args.size == 3) { "Invalid amount of arguments. 3 is required" }

			function.invoke(
				r,
				args[0] as? A ?: throw ClassCastException("Argument at 0 is supposed to be ${A::class.qualifiedName}, ${args[0]?.let { it::class.qualifiedName }} given"),
				args[1] as? B ?: throw ClassCastException("Argument at 1 is supposed to be ${B::class.qualifiedName}, ${args[1]?.let { it::class.qualifiedName }} given"),
				args[2] as? C ?: throw ClassCastException("Argument at 2 is supposed to be ${B::class.qualifiedName}, ${args[2]?.let { it::class.qualifiedName }} given"),
			)
		})
	}

	fun get(id: ResourceLocation): Function<R, T>? {
		synchronized(map) {
			return map[id]
		}
	}

	fun fromNetwork(buff: FriendlyByteBuf): BoundFunction<R, T>? {
		val id = ResourceLocation(buff.readUtf())
		val stream = DataInputStream(ByteBufInputStream(buff))

		val arguments = LinkedList<Any?>()

		for (i in 0 until stream.readVarIntLE()) {
			arguments.add(stream.readType())
		}

		return map[id]?.bind(arguments)
	}

	fun fromJson(value: JsonElement): BoundFunction<R, T>? {
		if (value !is JsonObject) {
			return null
		}

		val id = value["function"]?.asString ?: return null
		val argumentString = value["arguments"]?.asString ?: return null
		val stream = DataInputStream(FastByteArrayInputStream(Base64.getDecoder().decode(argumentString)))
		val arguments = LinkedList<Any?>()

		for (i in 0 until stream.readVarIntLE()) {
			arguments.add(stream.readType())
		}

		return map[ResourceLocation(id)]?.bind(arguments)
	}

	override fun serialize(src: BoundFunction<R, T>, typeOfSrc: Type, context: JsonSerializationContext): JsonElement {
		return src.toJson()
	}

	override fun deserialize(
		json: JsonElement,
		typeOfT: Type,
		context: JsonDeserializationContext
	): BoundFunction<R, T> {
		return fromJson(json) ?: throw JsonSyntaxException("Function is invalid")
	}

	override fun write(out: JsonWriter, value: BoundFunction<R, T>) {
		TypeAdapters.JSON_ELEMENT.write(out, value.toJson())
	}

	override fun read(`in`: JsonReader): BoundFunction<R, T> {
		return fromJson(TypeAdapters.JSON_ELEMENT.read(`in`)) ?: throw JsonSyntaxException("Function is invalid")
	}
}
