package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.Streams
import com.google.gson.GsonBuilder
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonSyntaxException
import com.mojang.blaze3d.platform.InputConstants
import com.mojang.brigadier.arguments.StringArgumentType
import com.mojang.brigadier.context.CommandContext
import it.unimi.dsi.fastutil.objects.Object2BooleanFunction
import it.unimi.dsi.fastutil.objects.Object2BooleanOpenHashMap
import it.unimi.dsi.fastutil.objects.Reference2BooleanFunction
import it.unimi.dsi.fastutil.objects.Reference2BooleanOpenHashMap
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import net.minecraft.ChatFormatting
import net.minecraft.commands.CommandSourceStack
import net.minecraft.commands.Commands
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.MutableComponent
import net.minecraft.resources.ResourceKey
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.MinecraftServer
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.server.packs.resources.SimpleJsonResourceReloadListener
import net.minecraft.tags.TagKey
import net.minecraft.util.profiling.ProfilerFiller
import net.minecraft.world.Container
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeType
import net.minecraft.world.level.ItemLike
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.event.AddReloadListenerEvent
import net.minecraftforge.event.OnDatapackSyncEvent
import net.minecraftforge.event.RegisterCommandsEvent
import net.minecraftforge.event.entity.player.ItemTooltipEvent
import net.minecraftforge.event.server.ServerStartedEvent
import net.minecraftforge.eventbus.api.IEventBus
import net.minecraftforge.fml.ModList
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.server.command.EnumArgument
import org.apache.logging.log4j.LogManager
import org.lwjgl.glfw.GLFW
import ru.dbotthepony.mc.otm.MINECRAFT_SERVER
import ru.dbotthepony.mc.otm.NULLABLE_MINECRAFT_SERVER
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.SERVER_IS_LIVE
import ru.dbotthepony.mc.otm.SystemTime
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.container.iterator
import ru.dbotthepony.mc.otm.container.spliterator
import ru.dbotthepony.mc.otm.container.stream
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.filterNotNull
import ru.dbotthepony.mc.otm.core.formatMatterFull
import ru.dbotthepony.mc.otm.core.formatSiComponent
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.core.isActuallyEmpty
import ru.dbotthepony.mc.otm.core.isZero
import ru.dbotthepony.mc.otm.core.orNull
import ru.dbotthepony.mc.otm.core.readItemType
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.core.stream
import ru.dbotthepony.mc.otm.core.writeItemType
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.RegistryNetworkChannel
import ru.dbotthepony.mc.otm.registry.RegistryDelegate
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper
import java.io.File
import java.math.BigInteger
import java.util.*
import java.util.function.Supplier
import java.util.stream.Stream
import kotlin.ConcurrentModificationException
import kotlin.collections.ArrayDeque
import kotlin.collections.ArrayList
import kotlin.collections.HashMap
import kotlin.collections.LinkedHashMap
import kotlin.math.pow

internal sealed class MutableEntry(
	val modificationChain: MutableList<ResourceLocation>,
	var matter: ImpreciseFraction,
	var complexity: Double,
	var priority: Int
) {
	abstract fun freeze(): Entry
}

internal class MutableTagEntry(
	val tag: TagKey<Item>,
	modificationChain: MutableList<ResourceLocation>,
	matter: ImpreciseFraction,
	complexity: Double,
	priority: Int
) : MutableEntry(modificationChain, matter, complexity, priority) {
	override fun freeze(): Entry {
		return TagEntry(tag, ImmutableList.copyOf(modificationChain), matter, complexity, priority)
	}
}

internal class MutableKeyEntry(
	val key: ResourceLocation,
	modificationChain: MutableList<ResourceLocation>,
	matter: ImpreciseFraction,
	complexity: Double,
	priority: Int
) : MutableEntry(modificationChain, matter, complexity, priority) {
	init {
		if (key == AIR) {
			throw JsonSyntaxException("you wot, can't modify $key")
		}
	}

	override fun freeze(): Entry {
		return KeyEntry(key, ImmutableList.copyOf(modificationChain), matter, complexity, priority)
	}

	companion object {
		private val AIR = ResourceLocation("minecraft", "air")
	}
}

internal sealed class Entry(
	val modificationChain: List<ResourceLocation>,
	final override val matter: ImpreciseFraction,
	final override val complexity: Double,
	val priority: Int,
) : IMatterValue {
	companion object : Entry(listOf(), ImpreciseFraction.ZERO, 0.0, Int.MIN_VALUE)
}

internal class TagEntry(
	val tag: TagKey<Item>,
	modificationChain: List<ResourceLocation>,
	matter: ImpreciseFraction,
	complexity: Double,
	priority: Int,
) : Entry(modificationChain, matter, complexity, priority) {
	val bound by lazy {
		val manager = ForgeRegistries.ITEMS.tags() ?: throw NullPointerException("Forge registry of Items has no tags!")
		manager.getTag(tag)
	}
}

internal class KeyEntry(
	val key: ResourceLocation,
	modificationChain: List<ResourceLocation>,
	matter: ImpreciseFraction,
	complexity: Double,
	priority: Int,
) : Entry(modificationChain, matter, complexity, priority)

private fun transformQuotes(it: String?): String = if (it != null) '"' + it.replace("\"", "\"\"") + '"' else "NULL"

object MatterManager {
	const val MATTER_DIRECTORY = "otm_matter"
	const val FINDER_DIRECTORY = "otm_recipe_finder"
	private val LOGGER = LogManager.getLogger()

	private object Registry : SimpleJsonResourceReloadListener(GsonBuilder().setPrettyPrinting().disableHtmlEscaping().create(), MATTER_DIRECTORY) {
		val keyEntries = HashMap<ResourceLocation, KeyEntry>()
		val tagEntries = LinkedHashMap<TagKey<Item>, TagEntry>()

		val keyEntriesBlacklist = HashMap<ResourceLocation, BlacklistAction>()
		val tagEntriesBlacklist = LinkedHashMap<TagKey<Item>, BlacklistAction>()

		val computeKeys = HashMap<ResourceLocation, ComputeAction>()
		val computeTags = HashMap<TagKey<Item>, ComputeAction>()

		var ready = false
			private set

		/**
		 * Get matter value for item defined by `insert` or `update` registry actions
		 */
		fun direct(value: Item): IMatterValue {
			check(ready) { "Not ready" }

			if (isBlacklisted(value))
				return IMatterValue.Companion

			val key = value.registryName ?: throw NullPointerException("$value has no registry name!")

			val keyEntry = keyEntries[key]

			if (keyEntry != null) {
				return keyEntry
			}

			val reverse = ForgeRegistries.ITEMS.tags()!!.getReverseTag(value).orElse(null)

			if (reverse != null) {
				return reverse.tagKeys
					.map { tagEntries[it] }
					.filterNotNull()
					.sorted { o1, o2 -> o1.priority.compareTo(o2.priority) }
					.findFirst()
					.orElse(null) ?: IMatterValue.Companion
			}

			return Entry.Companion
		}

		/**
		 * Get matter value for item defined by `compute` registry action
		 */
		fun indirect(value: Item): Result {
			check(ready) { "Not ready" }

			if (isBlacklisted(value))
				return Result.MISSING

			var compute = computeKeys[value.registryName!!]

			if (compute == null) {
				val reverse = ForgeRegistries.ITEMS.tags()!!.getReverseTag(value).orElse(null)

				if (reverse != null) {
					compute = reverse.tagKeys
						.map { computeTags[it] }
						.filterNotNull()
						.sorted()
						.findFirst()
						.orElse(null)
				}
			}

			if (compute != null) {
				return compute.tryCompute()
			}

			return Result.MISSING
		}

		private val blacklistCacheKey = Reference2BooleanOpenHashMap<Item>()
		private val blacklistCacheTag = Object2BooleanOpenHashMap<TagKey<Item>>() // despite them being interned as of 1.19.2, i consider this internal detail

		fun isBlacklisted(value: Item): Boolean {
			return blacklistCacheKey.computeIfAbsent(value, Reference2BooleanFunction fn@{
				val key = value.registryName ?: throw NullPointerException("$value has no registry name!")

				if (key in keyEntriesBlacklist)
					return@fn true

				val reverse = ForgeRegistries.ITEMS.tags()!!.getReverseTag(value).orElse(null)

				if (reverse != null) {
					return@fn reverse.tagKeys.anyMatch { it in tagEntriesBlacklist }
				}

				return@fn false
			})
		}

		fun isBlacklisted(value: TagKey<Item>): Boolean {
			return blacklistCacheTag.computeIfAbsent(value, Object2BooleanFunction fn@{
				return@fn value in tagEntriesBlacklist || ForgeRegistries.ITEMS.tags()!!.getTag(value).let {
					if (it.isEmpty || !it.isBound) {
						false
					} else {
						it.stream().anyMatch { item -> !isBlacklisted(item) }
					}
				}
			})
		}

		override fun prepare(resourceManager: ResourceManager, profilerFiller: ProfilerFiller): MutableMap<ResourceLocation, JsonElement> {
			ready = false
			registryIsBuilt = false
			val time = System.nanoTime()
			LOGGER.info("Slicing matter values...")
			profilerFiller.push("otm_matter_registry_prepare")

			try {
				val prep = super.prepare(resourceManager, profilerFiller)
				val diff = System.nanoTime() - time
				LOGGER.info("Slicing matter values took ${diff / 1_000_000L}ms")
				return prep
			} finally {
				profilerFiller.pop()
			}
		}

		override fun apply(map: MutableMap<ResourceLocation, JsonElement>, resourceManager: ResourceManager, profilerFiller: ProfilerFiller) {
			registryIsBuilt = false
			val time = System.nanoTime()
			LOGGER.info("Stitching matter values")
			profilerFiller.push("otm_matter_registry")

			try {
				keyEntries.clear()
				tagEntries.clear()

				computeKeys.clear()
				computeTags.clear()

				blacklistCacheKey.clear()
				blacklistCacheTag.clear()

				keyEntriesBlacklist.clear()
				tagEntriesBlacklist.clear()

				val add = ArrayList<Pair<InsertAction, ResourceLocation>>()
				val addOrReplace = ArrayList<Pair<InsertAction, ResourceLocation>>()
				val update = ArrayList<Pair<UpdateAction, ResourceLocation>>()
				val delete = ArrayList<Pair<DeleteAction, ResourceLocation>>()

				for ((key, json) in map) {
					if (json !is JsonObject) {
						throw JsonParseException("Matter value $key has invalid type: ${json::class.qualifiedName} ($json)")
					}

					when (val action = json["action"]?.asString?.lowercase() ?: throw JsonParseException("Missing `action` json element in $key. Possible values are: (INSERT, DELETE and UPDATE (case-insensitive))")) {
						"insert" -> {
							val value = InsertAction(json)

							if (value.replaceIfExists) {
								addOrReplace.add(value to key)
							} else {
								add.add(value to key)
							}
						}

						"update" -> update.add(UpdateAction(json) to key)
						"delete" -> delete.add(DeleteAction(json) to key)

						"compute" -> {
							try {
								ComputeAction(json).also {
									if (it.key != null) {
										computeKeys.compute(it.key) { _, existing -> if (it.canReplace(existing)) it else existing }
									} else {
										computeTags.compute(it.tag!!) { _, existing -> if (it.canReplace(existing)) it else existing }
									}
								}
							} catch (err: Throwable) {
								throw JsonSyntaxException("Processing $key failed", err)
							}
						}

						"blacklist" -> {
							val value = BlacklistAction(json)

							if (value.key != null) {
								keyEntriesBlacklist[value.key] = value
							} else {
								tagEntriesBlacklist[value.tag!!] = value
							}

							delete.add(value to key)
						}

						else -> throw JsonParseException("Unknown action $action of $key")
					}
				}

				val registry = ArrayList<MutableEntry>()

				for ((action, key) in add) action.update(registry, key)
				for ((action, key) in addOrReplace) action.update(registry, key)
				for ((action, key) in delete) action.update(registry, key)
				for ((action, key) in update) action.update(registry, key)

				for (key in keyEntriesBlacklist.keys) computeKeys.remove(key)
				for (key in tagEntriesBlacklist.keys) computeTags.remove(key)

				val tags = HashMap<TagKey<Item>, MutableTagEntry>()
				val keys = HashMap<ResourceLocation, MutableKeyEntry>()

				for (entry in registry) {
					if (entry is MutableTagEntry) {
						val existing = tags[entry.tag]

						if (existing == null || existing.priority < entry.priority) {
							tags[entry.tag] = entry
						}
					} else if (entry is MutableKeyEntry) {
						val existing = keys[entry.key]

						if (existing == null || existing.priority < entry.priority) {
							keys[entry.key] = entry
						}
					}
				}

				for (key in keys.values) {
					keyEntries[key.key] = key.freeze() as KeyEntry
				}

				val tagEntries = ArrayList<TagEntry>()

				for (tag in tags.values) {
					tagEntries.add(tag.freeze() as TagEntry)
				}

				tagEntries.sortBy { it.priority }

				for (entry in tagEntries) {
					this.tagEntries[entry.tag] = entry
				}
			} finally {
				profilerFiller.pop()
			}

			val diff = System.nanoTime() - time
			LOGGER.info("Stitching matter values took ${diff / 1_000_000L}ms")

			ready = true
			finishUpIfRequiredAndPossible(NULLABLE_MINECRAFT_SERVER)
		}
	}

	internal data class Result(val type: Type, val value: IMatterValue? = null) {
		constructor(value: IMatterValue) : this(Type.RESOLVED, value)

		val isSkipped get() = type === Type.SKIPPED
		val isMissing get() = type === Type.MISSING

		enum class Type {
			RESOLVED, SKIPPED, MISSING
		}

		companion object {
			// recursive recipes should be ignored until we resolve everything else
			val SKIPPED = Result(Type.SKIPPED)

			// missing matter values are fatal, whole recipe chain is then considered defunct
			val MISSING = Result(Type.MISSING)
		}
	}

	private object Resolver : SimpleJsonResourceReloadListener(GsonBuilder().setPrettyPrinting().disableHtmlEscaping().create(), FINDER_DIRECTORY) {
		val delegate = RegistryDelegate<Finder>("recipe_finder") {
			disableSync()
			disableSaving()
		}

		var ready = false
			private set

		val registrar: DeferredRegister<Finder> = DeferredRegister.create(delegate.key, OverdriveThatMatters.MOD_ID)

		init {
			registrar.register("simple") {
				Finder { server, data ->
					val location = (data["recipe_type"] ?: throw JsonSyntaxException("Missing recipe type")).let { ResourceLocation.tryParse(it.asString) } ?: throw JsonSyntaxException("Invalid recipe type: ${data["recipe_type"]}")

					if (!ForgeRegistries.RECIPE_TYPES.containsKey(location)) {
						LOGGER.error("Invalid or missing recipe category: $location!")
						return@Finder Stream.empty()
					}

					val findRecipeType = ForgeRegistries.RECIPE_TYPES.getValue(location) as RecipeType<Recipe<Container>>? ?: throw ConcurrentModificationException()

					val isCritical = data["is_critical"]?.asBoolean ?: true
					val ignoreDamageables = data["ignore_damageables"]?.asBoolean ?: false
					val allowBacktrack = data["allow_backtrack"]?.asBoolean ?: true

					var stream = server.recipeManager.byType(findRecipeType).values.parallelStream().filter { !it.isIncomplete }

					if (ignoreDamageables) {
						stream = stream.filter { it.ingredients.stream().flatMap { it.items.stream() }.noneMatch { it.isDamageableItem } }
					}

					stream.map {
						ResolvedRecipe(
							it.ingredients.stream()
								.filter { !it.isActuallyEmpty }
								.map { it.items.stream().map(::ImmutableStack) },
							ImmutableStack(it.resultItem),
							isCritical = isCritical,
							name = it.id,
							allowBacktrack = allowBacktrack
						)
					}
					.filter {
						if (it.inputs.isEmpty()) {
							LOGGER.warn("${it.formattedName} with output '${it.output.item.registryName}' ended up with no inputs!")
							false
						} else {
							true
						}
					}
				}
			}
		}

		private var foundResolvers: Map<ResourceLocation, Pair<Finder, JsonObject>> = ImmutableMap.of()

		override fun prepare(
			p_10771_: ResourceManager,
			p_10772_: ProfilerFiller
		): MutableMap<ResourceLocation, JsonElement> {
			registryIsBuilt = false
			return super.prepare(p_10771_, p_10772_)
		}

		override fun apply(values: Map<ResourceLocation, JsonElement>, resourceManager: ResourceManager, profilerFiller: ProfilerFiller) {
			registryIsBuilt = false
			determinedValues.clear()
			commentary.clear()

			val builder = ImmutableMap.Builder<ResourceLocation, Pair<Finder, JsonObject>>()

			for ((key, json) in values) {
				if (json !is JsonObject) {
					throw JsonParseException("$key is not a json object")
				}

				val location = (json["type"] ?: throw JsonSyntaxException("Missing resolver type")).let { ResourceLocation.tryParse(it.asString) } ?: throw JsonSyntaxException("Invalid resolver type: ${json["type"]}")

				if (!recipeFinders.containsKey(location)) {
					throw JsonParseException("Resolver type $location does not exist (in $key)")
				}

				val resolver = recipeFinders.getValue(location) ?: throw ConcurrentModificationException()
				builder.put(key, resolver to json)
			}

			foundResolvers = builder.build()
			ready = true
			finishUpIfRequiredAndPossible(NULLABLE_MINECRAFT_SERVER)
		}

		private data class Accumulator(
			val input2Recipes: Reference2ObjectOpenHashMap<Item, ReferenceOpenHashSet<ResolvedRecipe>> = Reference2ObjectOpenHashMap(),
			val output2Recipes: Reference2ObjectOpenHashMap<Item, ReferenceOpenHashSet<ResolvedRecipe>> = Reference2ObjectOpenHashMap(),
		) {
			val heuristicsSize: Long
				get() {
					return input2Recipes.size.toLong() + output2Recipes.size.toLong()
				}

			fun combine(other: Accumulator) {
				for ((k, v) in other.input2Recipes) {
					val existing = input2Recipes[k]

					if (existing == null) {
						input2Recipes[k] = v
					} else {
						existing.addAll(v)
					}
				}

				for ((k, v) in other.output2Recipes) {
					val existing = output2Recipes[k]

					if (existing == null) {
						output2Recipes[k] = v
					} else {
						existing.addAll(v)
					}
				}
			}

			fun add(recipe: ResolvedRecipe) {
				output2Recipes.computeIfAbsent(recipe.output.item, Reference2ObjectFunction { ReferenceOpenHashSet() }).add(recipe)

				for (input1 in recipe.inputs) {
					for (input in input1) {
						input2Recipes.computeIfAbsent(input.item, Reference2ObjectFunction { ReferenceOpenHashSet() }).add(recipe)
					}
				}
			}
		}

		private var input2Recipes: Map<Item, Collection<ResolvedRecipe>> = mapOf()
		private var output2Recipes: Map<Item, Collection<ResolvedRecipe>> = mapOf()

		private val determinedValues = Reference2ObjectOpenHashMap<Item, IMatterValue>()
		private val seenItems = ArrayDeque<Item>()

		private var changes = false
		private var cachedIterationResults = Reference2ObjectOpenHashMap<Item, Result>()

		private fun doTryToBacktrack(item: Item, makeCommentary: Boolean): Result {
			val recipes = input2Recipes[item]

			if (recipes == null || recipes.isEmpty() || !recipes.all { it.allowBacktrack } || !recipes.all { it.inputs.all { it.all { it.item == item } } }) {
				if (makeCommentary)
					comment(item, TextComponent("Item '${item.registryName}' has no recipes"))

				return Result.MISSING
			}

			var minimal: IMatterValue? = null
			var minimalMultiplier = 0.0
			var foundRecipe: ResolvedRecipe? = null

			for (recipe in recipes) {
				val value = doDetermineValue(recipe.output.item)

				if (value.value != null) {
					if (minimal == null || minimal < value.value) {
						minimal = value.value
						minimalMultiplier = recipe.output.multiplier / recipe.inputs.size
						foundRecipe = recipe
					}
				} else if (!value.isSkipped) {
					LOGGER.error("${item.registryName} cant resolve ${recipe.output.item}")
					minimal = null
					break
				}
			}

			if (minimal == null) {
				if (makeCommentary)
					comment(item, TextComponent("Item '${item.registryName}' has no valid backtracking recipes, and no output recipes"))

				return Result.MISSING
			}

			val result = MatterValue(minimal.matter * minimalMultiplier, minimal.complexity * minimalMultiplier)

			changes = true
			determinedValues[item] = result
			comment(item, TextComponent("Matter value backtracked from ${foundRecipe!!.formattedName} at iteration $iteration"))
			return Result(result)
		}

		private fun tryToBacktrack(item: Item, asFallback: Boolean): Result {
			val getResult = cachedIterationResults[item]

			if (getResult != null) {
				return getResult
			}

			if (Registry.isBlacklisted(item))
				return Result.MISSING

			val value = Registry.direct(item)

			if (value.hasMatterValue)
				return Result(value)

			val value2 = determinedValues[item]
			if (value2?.hasMatterValue == true)
				return Result(value2)

			if (item in seenItems)
				if (!asFallback || item == seenItems.last() && seenItems.count { it == item } == 1)
					return Result.SKIPPED

			seenItems.addLast(item)

			try {
				val value3 = Registry.indirect(item)
				if (value3.value != null)
					return value3

				val result = doTryToBacktrack(item, !asFallback)
				cachedIterationResults[item] = result
				return result
			} finally {
				seenItems.removeLast()
			}
		}

		private fun doDetermineValue(item: Item): Result {
			var minimalMatter: ImpreciseFraction? = null
			var minimalComplexity: Double? = null

			val recipes = output2Recipes[item]

			if (recipes == null || recipes.isEmpty()) {
				return tryToBacktrack(item, false)
			}

			var hadSkips = false
			var minimalRecipe: ResolvedRecipe? = null

			recipesLoop@ for (recipe in recipes) {
				if (recipe.inputs.isEmpty()) {
					// TODO: should we ignore empty recipes?
					comment(item, TextComponent("${recipe.formattedName} is empty"))

					continue
				}

				var accumulatedMatter: ImpreciseFraction? = null
				var accumulatedComplexity: Double? = null

				inputsLoop@ for ((i, inputs) in recipe.inputs.withIndex()) {
					var minimal: IMatterValue? = null
					var minimalMultiplier = 0.0

					innerInputsLoop@ for (input in inputs) {
						val ivalue = determineValue(input.item)

						if (ivalue.isMissing) {
							comment(item, TextComponent("Input '${input.item.registryName}' at input slot $i in ${recipe.formattedName} has no matter value"))

							if (recipe.isCritical) {
								return Result.MISSING
							} else {
								continue@recipesLoop
							}
						} else if (ivalue.isSkipped) {
							comment(item, TextComponent("Input '${input.item.registryName}' at input slot $i in ${recipe.formattedName} is recursive"))

							if (inputs.size == 1) {
								hadSkips = true
								continue@recipesLoop
							} else {
								continue@innerInputsLoop
							}
						}

						if (minimal == null || minimal > ivalue.value!!) {
							minimal = ivalue.value
							minimalMultiplier = input.multiplier
						}
					}

					if (minimal == null || !minimal.hasMatterValue) {
						comment(item, TextComponent("'${recipe.formattedName}' has invalid input at slot $i (possible inputs: ${inputs.joinToString(", ", transform = { it.item.registryName.toString() }) }) (???)"))
						return Result.MISSING
					}

					if (accumulatedMatter == null || accumulatedComplexity == null) {
						accumulatedMatter = minimal.matter * minimalMultiplier
						accumulatedComplexity = minimal.complexity * minimalMultiplier
					} else {
						accumulatedMatter += minimal.matter * minimalMultiplier
						accumulatedComplexity += minimal.complexity * minimalMultiplier
					}
				}

				if (accumulatedMatter == null || accumulatedComplexity == null) {
					throw RuntimeException("This piece should be unreachable")
				}

				if (!accumulatedMatter.isPositive || accumulatedComplexity <= 0.0) {
					comment(item, TextComponent("${recipe.formattedName} ended up with negative matter value and/or complexity (???)"))
					return Result.MISSING
				}

				if (minimalMatter == null || minimalMatter > accumulatedMatter) {
					minimalMatter = recipe.transformMatterValue(accumulatedMatter * recipe.output.multiplier)
					minimalComplexity = recipe.transformComplexity(accumulatedComplexity * recipe.output.multiplier)
					minimalRecipe = recipe
				}
			}

			if (minimalMatter == null || minimalComplexity == null) {
				if (hadSkips) {
					val backtrack = tryToBacktrack(item, true)

					if (backtrack.value == null) {
						return Result.SKIPPED
					} else {
						return backtrack
					}
				}

				comment(item, TextComponent("'${item.registryName}' ended up with no valid recipes (???)"))

				return Result.MISSING
			}

			val result = MatterValue(minimalMatter, minimalComplexity)

			changes = true
			determinedValues[item] = result
			comment(item, TextComponent("Matter value derived from ${minimalRecipe!!.formattedName} at iteration $iteration"))

			return Result(result)
		}

		fun determineValue(item: Item): Result {
			val getResult = cachedIterationResults[item]

			if (getResult != null)
				return getResult

			if (Registry.isBlacklisted(item))
				return Result.MISSING

			val direct = Registry.direct(item)
			if (direct.hasMatterValue)
				return Result(direct)

			val determined = determinedValues[item]
			if (determined?.hasMatterValue == true)
				return Result(determined)

			if (item in seenItems)
				return Result.SKIPPED

			seenItems.addLast(item)

			try {
				val indirect = Registry.indirect(item)
				if (indirect.value != null)
					return indirect

				val result = doDetermineValue(item)
				cachedIterationResults[item] = result
				return result
			} finally {
				seenItems.removeLast()
			}
		}

		fun resolve(server: MinecraftServer) {
			var time = SystemTime()
			LOGGER.info("Finding recipes...")

			determinedValues.clear()
			seenItems.clear()

			val foundStreams = foundResolvers.map {
				try {
					it.value.first.find(server, it.value.second)
				} catch(err: Throwable) {
					LOGGER.fatal("Recipe resolver ${it.key} experienced internal error", err)
					throw RuntimeException("Recipe resolver ${it.key} experienced internal error", err)
				}
			}

			val sequentialStreams = foundStreams.filter { !it.isParallel }
			val parallelStreams = foundStreams.filter { it.isParallel }

			val streamA = Streams.concat(*sequentialStreams.toTypedArray())
			val streamB = Streams.concat(*parallelStreams.toTypedArray())

			val resultA = streamA.collect(::Accumulator, Accumulator::add, Accumulator::combine)
			val resultB = streamB.collect(::Accumulator, Accumulator::add, Accumulator::combine)

			val result: Accumulator

			if (resultA.heuristicsSize > resultB.heuristicsSize) {
				result = resultA
				resultA.combine(resultB)
			} else {
				result = resultB
				resultB.combine(resultA)
			}

			val (input2Recipes, output2Recipes) = result

			LOGGER.info("Finding recipes took ${time.millis}ms, involving ${input2Recipes.keys.size} unique inputs and ${output2Recipes.keys.size} unique outputs")
			LOGGER.info("Resolving recipes...")

			this.input2Recipes = input2Recipes
			this.output2Recipes = output2Recipes

			changes = true
			iteration = 0
			var ops = 0

			val toDetermine = ReferenceLinkedOpenHashSet<Item>()
			toDetermine.addAll(input2Recipes.keys)
			toDetermine.addAll(output2Recipes.keys)

			if (LOGGER.isDebugEnabled) {
				LOGGER.debug("Gonna try to search for matter values for next items:")

				val names = ArrayList<String>()
				var length = 0

				for (value in toDetermine) {
					val name = value.registryName!!.toString()

					if (length != 0 && length + name.length < 400) {
						names.add(name)
						length += name.length
					} else if (length == 0) {
						names.add(name)
						length = name.length
					} else {
						LOGGER.debug(names.joinToString(", "))
						names.clear()
						length = 0
					}
				}

				if (names.isNotEmpty()) {
					LOGGER.debug(names.joinToString(", "))
				}
			}

			time = SystemTime()

			while (changes) {
				ops += cachedIterationResults.size
				cachedIterationResults = Reference2ObjectOpenHashMap()
				changes = false
				iteration++

				val iterator = toDetermine.iterator()

				for (value in iterator) {
					if (determineValue(value).value?.hasMatterValue == true) {
						iterator.remove()
					}
				}
			}

			for (item in toDetermine) {
				if (item !in commentary || commentary[item]!!.isEmpty()) {
					if (Registry.isBlacklisted(item)) {
						comment(item, TextComponent("Item ${item.registryName} is blacklisted from having matter value"))
					} else {
						comment(item, TextComponent("Item ${item.registryName} was never visited"))
					}
				}
			}

			LOGGER.info("Resolving recipes took ${time.millis}ms in $iteration iterations with ~$ops operations, determined ${determinedValues.size} matter values")
		}
	}

	var iteration = 0

	internal fun compute(value: Item): Result {
		return Resolver.determineValue(value)
	}

	private val seenTags = ArrayDeque<TagKey<Item>>()

	internal fun compute(value: TagKey<Item>): Result {
		if (value in seenTags) {
			return Result.SKIPPED
		}

		seenTags.addLast(value)

		try {
			val result = Registry.tagEntries[value]

			if (result != null) {
				return Result(result)
			}

			val computed = Registry.computeTags[value]

			if (computed != null) {
				return computed.tryCompute()
			}

			return Result.MISSING
		} finally {
			seenTags.removeLast()
		}
	}

	@JvmStatic
	fun hasMatterValue(value: Item): Boolean {
		return get(value).hasMatterValue
	}

	private fun get(value: ItemStack, level: Int, accountForStackSize: Boolean): IMatterValue {
		if (value.isEmpty) {
			return IMatterValue.Companion
		}

		var matter: IMatterValue
		var undamagedMatterValue: IMatterValue

		if (value.item is IMatterItem)
			matter = (value.item as IMatterItem).getMatterValue(value) ?: IMatterValue.Companion
		else
			matter = get(value.item)

		if (matter.hasMatterValue && value.isDamageableItem) {
			val severity = 1.0 - value.damageValue.toDouble() / value.maxDamage.toDouble()
			undamagedMatterValue = matter
			matter = MatterValue(matter.matter * severity, matter.complexity * (0.5 + severity / 2))
		} else {
			undamagedMatterValue = matter
		}

		val matterCap = value.getCapability(MatteryCapability.MATTER).orNull()

		if (matterCap != null) {
			matter = MatterValue(matter.matter + matterCap.storedMatter, matter.complexity)
			undamagedMatterValue = MatterValue(undamagedMatterValue.matter + matterCap.storedMatter, undamagedMatterValue.complexity)
		}

		val drive = value.getCapability(MatteryCapability.DRIVE).orNull()

		if (drive != null && drive.storageType === OverdriveThatMatters.INSTANCE.ITEM_STORAGE()) {
			for (item in (drive as IMatteryDrive<ItemStackWrapper>).stacks) {
				val tuple = get(item.stack.stack, level + 1, true)

				if (tuple.hasMatterValue) {
					matter += tuple
					undamagedMatterValue += tuple
				}
			}
		}

		value.getCapability(ForgeCapabilities.ITEM_HANDLER).ifPresentK {
			it.stream().filter { !it.isEmpty }.map { get(it, level + 1, true) }.reduce { a, b ->
				if (a.hasMatterValue && b.hasMatterValue) {
					return@reduce MatterValue(a.matter + b.matter, a.complexity + b.complexity)
				} else if (a.hasMatterValue) {
					return@reduce a
				} else if (b.hasMatterValue) {
					return@reduce b
				} else {
					return@reduce IMatterValue.Companion
				}
			}.ifPresent {
				matter += it
				undamagedMatterValue += it
			}
		}

		if (matter.hasMatterValue && value.count > 1 && accountForStackSize) {
			return MatterValue(matter.matter + undamagedMatterValue.matter * (value.count - 1), matter.complexity + undamagedMatterValue.complexity * (value.count - 1))
		}

		return matter
	}

	/**
	 * Matter value for itemstack, accounting for capabilities and stack size
	 */
	@JvmStatic
	@JvmOverloads
	fun get(value: ItemStack, accountForStackSize: Boolean = true): IMatterValue {
		return get(value, 0, accountForStackSize)
	}

	@JvmStatic
	fun hasMatterValue(value: ItemStack): Boolean {
		return get(value).hasMatterValue
	}

	@JvmStatic
	fun canDecompose(value: Item): Boolean {
		if (value is IMatterItem)
			return value.canDecompose(ItemStack.EMPTY)

		return get(value).hasMatterValue
	}

	@JvmStatic
	fun canDecompose(value: ItemStack): Boolean {
		var can = true

		if (value.item is IMatterItem)
			can = (value.item as IMatterItem).canDecompose(value)

		can = can && (value.getCapability(MatteryCapability.MATTER).orNull()?.storedMatter ?: ImpreciseFraction.ZERO).isZero
		can = can && (value.getCapability(MatteryCapability.DRIVE).orNull()?.storedCount ?: BigInteger.ZERO).isZero

		return can && get(value).hasMatterValue
	}

	@JvmStatic
	fun getResearchAdvance(complexity: Double): Double {
		if (complexity <= 0.0) {
			return 0.0
		}

		return (1.0 / (complexity.pow(1.0 / 3.0) * 12.0).coerceAtLeast(0.01)).coerceAtMost(0.2).coerceAtLeast(0.025)
	}

	@JvmStatic
	fun getResearchAdvance(stack: ItemStack): Double {
		return getResearchAdvance(get(stack).complexity)
	}

	@JvmStatic
	fun getResearchAdvance(stack: ItemLike): Double {
		return getResearchAdvance(this@MatterManager.get(stack.asItem()).complexity)
	}

	fun tooltipEvent(event: ItemTooltipEvent) {
		val window = minecraft.window.window

		if (InputConstants.isKeyDown(window, GLFW.GLFW_KEY_LEFT_SHIFT) || InputConstants.isKeyDown(window, GLFW.GLFW_KEY_RIGHT_SHIFT)) {
			val matter = get(event.itemStack)

			if (matter.hasMatterValue) {
				if (matter.complexity >= 1.0E-3) {
					event.toolTip.add(
						TranslatableComponent("otm.gui.matter.format_and_complexity",
							matter.matter.formatMatterFull(),
							matter.complexity.formatSiComponent(TranslatableComponent("otm.gui.ticks"))
						).withStyle(ChatFormatting.AQUA))
				} else {
					event.toolTip.add(matter.matter.formatMatterFull().withStyle(ChatFormatting.AQUA))
				}
			}

			if (event.flags.isAdvanced) {
				val commentary = getCommentary(event.itemStack.item)

				if (commentary != null) {
					for (value in commentary) {
						event.toolTip.add(value.withStyle(ChatFormatting.DARK_GRAY))
					}
				}
			}
		}
	}

	/**
	 * because it is immutable
	 */
	data class ImmutableStack(
		val item: Item,
		val multiplier: Double = 1.0
	) {
		constructor(item: Item, count: Int) : this(item, 1.0 / count.toDouble())
		constructor(item: ItemStack) : this(item.item, item.count)
	}

	class ResolvedRecipe(
		inputs: Stream<Stream<ImmutableStack>>,
		val output: ImmutableStack,
		val transformMatterValue: (ImpreciseFraction) -> ImpreciseFraction = { it },
		val transformComplexity: (Double) -> Double = { it },

		/**
		 * if we can't resolve one of ingredient's matter value - consider output also having no matter value
		 */
		val isCritical: Boolean = true,
		val name: ResourceLocation? = null,

		/**
		 * Whenever to allow "backtracking" recipe ingredients to recipe outputs.
		 *
		 * When an item has no recipes, but takes part in recipes with output of already determined matter value
		 * (e.g. 4 rabbit hides -> 1 leather), we can assume that we can determine matter value of ingredients based on matter value of output
		 *
		 * "Backtrack" can happen if everything of the next holds true:
		 * * All recipes with item in question contains only that item; and
		 * * All recipes allow backtracking
		 */
		val allowBacktrack: Boolean = true,
	) {
		val inputs: List<List<ImmutableStack>> = inputs
			.map { it.filter { it.multiplier > 0.0 }.collect(ImmutableList.toImmutableList()) }
			.filter { it.isNotEmpty() }
			.collect(ImmutableList.toImmutableList())

		constructor(
			inputs: Collection<Collection<ImmutableStack>>,
			output: ImmutableStack,
			transformMatterValue: (ImpreciseFraction) -> ImpreciseFraction = { it },
			transformComplexity: (Double) -> Double = { it },
		) : this(inputs.stream().map { it.stream() }, output, transformMatterValue, transformComplexity)

		val formattedName: String
			get() = if (name == null) "One of recipes" else "Recipe '$name'"
	}

	fun interface Finder {
		/**
		 * Returned stream can be either parallel or sequential
		 *
		 * Parallel streams will _greatly_ improve performance on mid-end range computers,
		 * but you need to make sure your stream has no side effects (such as binding late-binding tags)
		 *
		 * Internally, OTM separate parallel streams from sequential so sequential streams happen strictly on server thread
		 * (if you can't avoid side effects, or don't bother with synchronization).
		 *
		 * You can safely call [MatterManager.getDirect] inside returned stream, both parallel and sequential.
		 */
		fun find(server: MinecraftServer, json: JsonObject): Stream<ResolvedRecipe>
	}

	/**
	 * Please use exclusively inside recipe finder
	 */
	@JvmStatic
	fun getDirect(value: Item): IMatterValue {
		return Registry.direct(value)
	}

	@JvmStatic val recipeFinders get() = Resolver.delegate.get()
	@JvmStatic val recipeFindersKey get() = Resolver.delegate.key

	private val commentary = Reference2ObjectOpenHashMap<Item, ArrayList<Component>>()

	@JvmStatic
	fun getCommentary(item: Item): List<MutableComponent>? {
		return commentary[item]?.let { source ->
			ArrayList<MutableComponent>(source.size)
				.also { for (value in source) it.add(value.copy()) }
		}
	}

	private fun comment(item: Item, comment: Component) {
		val target = commentary.computeIfAbsent(item, Reference2ObjectFunction { ArrayList(1) })

		if (comment !in target) {
			target.add(comment)
		}
	}

	fun reloadEvent(event: AddReloadListenerEvent) {
		event.addListener(Registry)
		event.addListener(Resolver)
	}

	fun initialize(bus: IEventBus) {
		bus.addListener(Resolver.delegate::build)
		Resolver.registrar.register(bus)
	}

	private enum class DumpFilter {
		ALL {
			override fun filter(input: Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>>): Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>> {
				return input
			}
		},
		ONLY_MISSING {
			override fun filter(input: Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>>): Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>> {
				return input.filter { !it.second.hasMatterValue }
			}
		},
		ONLY_VALID {
			override fun filter(input: Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>>): Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>> {
				return input.filter { it.second.hasMatterValue }
			}
		},
		ONLY_BLACKLISTED {
			override fun filter(input: Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>>): Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>> {
				return input.filter { Registry.isBlacklisted(it.first.value) }
			}
		},
		;

		abstract fun filter(input: Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>>): Stream<Pair<Map.Entry<ResourceKey<Item>, Item>, IMatterValue>>
	}

	private fun dumpRegistry(stack: CommandContext<CommandSourceStack>, filter: DumpFilter = DumpFilter.ALL, mod: String? = null): Int {
		val targetFile = File(MINECRAFT_SERVER.serverDirectory, "otm/registry_dumps/${filter.name.lowercase()}_${System.currentTimeMillis() / 1_000L}.csv")
		File(MINECRAFT_SERVER.serverDirectory, "otm/registry_dumps").mkdirs()

		stack.source.sendSuccess(TranslatableComponent("otm.dumping_matter_registry", targetFile.absolutePath), false)

		val writer = targetFile.bufferedWriter(Charsets.UTF_8)

		writer.write(arrayOf("Registry ID", "Matter Value", "Complexity", "Commentary").joinToString(";", transform = ::transformQuotes))
		writer.write("\n")

		var stream = ForgeRegistries.ITEMS.entries.stream()

		if (mod != null) {
			stream = stream.filter { it.key.location().namespace == mod }
		}

		if (filter != DumpFilter.ALL && filter != DumpFilter.ONLY_BLACKLISTED) {
			stream = stream.filter { !Registry.isBlacklisted(it.value) }
		}

		val stream2 = stream.map { it to get(it.value) }

		for ((entry, value) in filter.filter(stream2)) {
			val (key, item) = entry
			val isBlacklisted: Boolean

			if (filter == DumpFilter.ONLY_BLACKLISTED) {
				isBlacklisted = true
			} else if (filter == DumpFilter.ALL) {
				isBlacklisted = Registry.isBlacklisted(item)
			} else {
				isBlacklisted = false
			}

			if (isBlacklisted) {
				writer.write(arrayOf(key.location().toString(), "---", "---", "Item is blacklisted from having matter value").joinToString(";", transform = ::transformQuotes))
			} else {
				if (!value.hasMatterValue) {
					writer.write(arrayOf(key.location().toString(), "", "", commentary[item]?.joinToString("\n", transform = { transformQuotes(it.string) }) ?: "").joinToString(";", transform = ::transformQuotes))
				} else {
					writer.write(arrayOf(key.location().toString(), value.matter.toString(), value.complexity.toString(), commentary[item]?.joinToString("\n", transform = { transformQuotes(it.string) }) ?: "").joinToString(";", transform = ::transformQuotes))
				}
			}

			writer.write("\n")
		}

		writer.close()

		stack.source.sendSuccess(TranslatableComponent("otm.dumped_matter_registry", targetFile.absolutePath), true)

		return 0
	}

	fun addCommands(event: RegisterCommandsEvent) {
		event.dispatcher.register(
			Commands.literal("dump_matter_registry")
				.requires { it.hasPermission(Commands.LEVEL_OWNERS) }
				.executes(::dumpRegistry)
				.then(
					Commands.argument("mode", EnumArgument.enumArgument(DumpFilter::class.java)).executes { dumpRegistry(it, it.getArgument("mode", DumpFilter::class.java)) }
						.then(Commands.argument("mod", StringArgumentType.string())
							.executes { dumpRegistry(it, it.getArgument("mode", DumpFilter::class.java), StringArgumentType.getString(it, "mod")) }
							.suggests { context, builder ->
								val startingWith = builder.input.substring(builder.start).lowercase()
								ModList.get().mods.stream()
									.filter { it.modId.startsWith(startingWith) }
									.filter { key -> ForgeRegistries.ITEMS.entries.stream()
										.filter { it.key.location() == key }
										.anyMatch { !Registry.isBlacklisted(it.value) } }
									.forEach { builder.suggest(it.modId, TextComponent(it.displayName)) }
								builder.buildFuture()
							})
				)
		)
	}

	private val matterValues = Reference2ObjectOpenHashMap<Item, IMatterValue>()

	val map: Map<Item, IMatterValue> = Collections.unmodifiableMap(matterValues)

	private var registryIsBuilt = false

	private fun finishUpIfRequiredAndPossible(server: MinecraftServer?) {
		if (server == null)
			return

		if (registryIsBuilt || !Resolver.ready || !Registry.ready)
			return

		registryIsBuilt = true

		commentary.clear()
		matterValues.clear()

		Resolver.resolve(server)

		for (item in ForgeRegistries.ITEMS) {
			val value = compute(item).value

			if (value?.hasMatterValue == true)
				matterValues[item] = value
		}

		RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), SyncPacket(matterValues, commentary))
	}

	fun onDataPackSync(event: OnDatapackSyncEvent) {
		if (!registryIsBuilt)
			return

		if (event.player == null) {
			RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), SyncPacket(matterValues, commentary))
		} else {
			RegistryNetworkChannel.send(PacketDistributor.PLAYER.with { event.player ?: throw ConcurrentModificationException() }, SyncPacket(matterValues, commentary))
		}
	}

	fun onServerStarted(event: ServerStartedEvent) {
		check(Resolver.ready) { "Recipe resolver is not ready somehow" }
		check(Registry.ready) { "Matter registry is not ready somehow" }
		finishUpIfRequiredAndPossible(event.server ?: throw NullPointerException("what."))
	}

	fun get(value: Item): IMatterValue {
		return matterValues[value] ?: IMatterValue.Companion
	}

	fun readSyncPacket(buff: FriendlyByteBuf): SyncPacket {
		LOGGER.info("Received matter registry packet, ${buff.readableBytes()} bytes in size")

		val time = SystemTime()

		val result = SyncPacket(
			buff.readMap(FriendlyByteBuf::readItemType, FriendlyByteBuf::readMatterValue),
			buff.readMap(FriendlyByteBuf::readItemType) { self -> self.readCollection(::ArrayList, FriendlyByteBuf::readComponent) }
		)

		LOGGER.debug("Reading matter registry packet took ${time.millis}ms")

		return result
	}

	class SyncPacket(
		val values: Map<Item, IMatterValue>,
		val comments: Map<Item, Collection<Component>>
	) : MatteryPacket {
		override fun write(buff: FriendlyByteBuf) {
			val time = SystemTime()
			buff.writeMap(values, FriendlyByteBuf::writeItemType, FriendlyByteBuf::writeMatterValue)
			buff.writeMap(comments, FriendlyByteBuf::writeItemType) { self, value -> self.writeCollection(value, FriendlyByteBuf::writeComponent) }
			LOGGER.debug("Encoding matter registry packet took ${time.millis}ms, written total ${buff.writerIndex() - 1} bytes")
		}

		override fun play(context: Supplier<NetworkEvent.Context>) {
			if (SERVER_IS_LIVE)
				return // singleplayer or LAN host

			val time = SystemTime()

			matterValues.clear()
			matterValues.putAll(values)

			commentary.clear()

			for ((k, v) in comments) {
				commentary[k] = ArrayList<Component>(v.size).also { it.addAll(v) }
			}

			LOGGER.debug("Updating matter registry from packet took ${time.millis}ms")
		}
	}
}
